/*
   Dialogos - automate network communication
   Copyright (C) 2002 Markus Triska

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

   Contact me at triska@gmx.at
*/


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <stdarg.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>

#include "node.h"
#include "network.h"
#include "dialogos.tab.h"
#include "grep.h"

extern int yyerror(char *);
extern int lineno;

using namespace std;


/* I have taken a lot of code from Thomas Niemann's excellent tutorial on
   lex and yacc, available from http://members.xoom.com/thomasn .

	Many thanks!

*/


map<string, nodeType *> procedures;

nodeType *con(int i)
{
	nodeType *p = (nodeType *) malloc(sizeof(conNodeType));
	if (p == NULL) {
		yyerror("out of memory");
	}
	p->type = typeCon;
	p->con.value = i;
	return p;
}

nodeType *id(char *s)
{
	nodeType *p = (nodeType *) malloc(sizeof(idNodeType));
	if (p == NULL) {
		yyerror("out of memory");
	}
	p->type = typeId;
	p->id.str = s;
	return p;
}

nodeType *str(char *s)
{
	nodeType *p = (nodeType *) malloc(sizeof(strNodeType));
	if (p == NULL) {
		yyerror("out of memory");
	}
	p->type = typeStr;
	p->str.str = s;
	return p;
}

nodeType *opr(int oper, int nops, ...)
{
	va_list ap;
	nodeType *p;
	size_t size;
	int i;

	size = sizeof(oprNodeType) + (nops - 1)*sizeof(nodeType *);
	if ( (p = (nodeType *) malloc(size)) == NULL) {
		yyerror("out of memory");
	}

	p->type = typeOpr;
	p->opr.oper = oper;
	p->opr.nops = nops;
	va_start(ap, nops);
	for (i = 0; i < nops; i++) {
		p->opr.op[i] = va_arg(ap, nodeType *);
	}
	va_end(ap);
	return p;
}

void freeNode(nodeType *p)
{
	if (!p) return;
	switch(p->type) {
	case typeStr:
		if (p->str.str) free(p->str.str);
		break;
	case typeId:
		if (p->id.str) free(p->id.str);
		break;
	case typeOpr:
		for (int i = 0; i < p->opr.nops; i++) {
			freeNode(p->opr.op[i]);
		}
		break;
	}
	free(p);
}

void freeProcedures()
{
	for (map<string, nodeType *>::iterator it = procedures.begin(); it != procedures.end(); it++) {
		pair<string, nodeType *> p = *it;
		freeNode(p.second);
	}
}

int ex(nodeType *);

void registerProcedure(char *procname, nodeType *p)
{
	if (procedures.find(procname) == procedures.end()) {
		procedures[procname] = p;
	} else {
		cout << "Error: Procedure " << procname << " already defined. Ignoring redefinition." << endl;
	}
}

int performProcedure(char *procname)
{
	map<string, nodeType *>::iterator it = 	procedures.find(procname);
	if (it == procedures.end()) {
		cout << "PERFORM: Procedure " << procname << " not found." << endl;
	} else {
		return ex(procedures[procname]);
	}
	return 0;
}


string substOneLevelVars(string s)
{
	char refdelim = (variables["REFDELIM"])[0];
	string::iterator beginvar = find(s.begin(), s.end(), refdelim);
	string::iterator firstafterref = s.begin();
	ostringstream outstring;
	while (beginvar != s.end()) {
		// String contains variable reference
		string::iterator endvar = find(beginvar + 1, s.end(), refdelim);
		if (endvar == s.end()) {
			cout << "Error: string on line " << lineno << " contains begin of variable reference, but no end." << endl;
			break;
		} else {
			string variable(beginvar + 1, endvar);
			string value = "none";
			if (variables.find(variable) == variables.end()) {
				cout << "Warning: variable " << variable << " not initialized" << endl;
			}
			value = variables[variable];
			outstring <<  string(firstafterref, beginvar) << value;
			firstafterref = endvar + 1;
			beginvar = find(firstafterref, s.end(), refdelim);
		}
	}
	outstring << string(firstafterref, beginvar);
	return outstring.str();
}

string substVars(string s)
{
	string retstring = s;
	string oldstring;
	do {
		oldstring = retstring;
		retstring = substOneLevelVars(retstring);
	} while (retstring != oldstring);
	return retstring;
}

void doVariableAssignment(nodeType *p)
{
	string variable = p->opr.op[0]->id.str;

	switch (p->opr.nops) {
	case 2: switch(p->opr.op[1]->type) {
		case typeStr:
			variables[variable] = p->opr.op[1]->str.str;
			break;
		case typeId:
			variables[variable] = variables[p->opr.op[1]->id.str];
			break;
		case typeCon:
		case typeOpr:
			int value = ex(p->opr.op[1]);
			char buf[20];
			sprintf(buf, "%d", value);
			variables[variable] = buf;
			break;
		}
		break;
	case 3: {
		char *file = p->opr.op[1]->str.str;
		int offset = p->opr.op[2]->con.value;
		ifstream infile(file);
		if (!infile) {
			cout << "Error opening file " << file << endl;
		} else {
			infile.seekg(offset - 1);
			string value;
			infile >> value;
			variables[variable] = value;
			infile.close();
		}
		}
		break;
	case 4: {
		char *file = p->opr.op[1]->str.str;
		int offset1 = p->opr.op[2]->con.value;
		if (p->opr.op[3] == NULL) {
			// read in numeric
			ifstream infile(file);
			if (!infile) {
				cout << "Error opening file " << file << endl;
			} else {
				infile.seekg(offset1 - 1);
				int value;
				infile >> value;
				ostringstream ost; ost << value;
				string svalue = ost.str();
				variables[variable] = svalue;
				infile.close();
			}
		} else {
			ifstream infile(file);
			if (!infile) {
				cout << "Error opening file " << file << endl;
			} else {
				infile.seekg(offset1 - 1);
				string value;
				char c, delim;
				int offset2, numbytes;
				switch (p->opr.op[3]->type) {
				case typeCon:
					offset2 = p->opr.op[3]->con.value;
					numbytes = offset2 - offset1;
					for (int i = 0; i <= numbytes; i++) {
						char c = infile.get();
						if (infile.eof()) break;
						value += c;
					}
					break;
				case typeStr:
					delim = p->opr.op[3]->str.str[0];
					while ( (c = infile.get()) != delim) {
						if (infile.eof()) break;
						value += c;
					}
				}
				variables[variable] = value;
				infile.close();
			}
		}
		}
		break;
	}
}


void doProcessFile(nodeType *p, bool linewise = false)
{
	string fname = substVars(p->opr.op[0]->str.str);
	replace(fname.begin(), fname.end(), '/', '_');

	ifstream wordlist(fname.c_str());
	if (!wordlist) {
		cout << "PROCESS: could not open file " << fname << endl;
		return;
	}
	string usevar = p->opr.op[1]->id.str;

	while (wordlist && !wordlist.eof()) {
		string kw;
		if (linewise) {
			char c;
			// Read one line (maybe only one [CR]LF)
			for (; ;) {
				c = wordlist.get();
				if (wordlist.eof()) break;
				if (c == '\n') break;
				kw += c;
			}
		} else {
			wordlist >> kw;
		}
		if (kw.size()) {
			variables[usevar] = kw;
			ex(p->opr.op[2]);
		}
	}
	wordlist.close();
}


int ex(nodeType *p)
{
	if (p->type == typeCon) return p->con.value;
	if (p->type == typeId)  return atoi(variables[p->id.str].c_str());

	switch (p->opr.oper) {
	case APPEND: {
		ofstream outfile(substVars(p->opr.op[1]->str.str).c_str(), ios::app);
		outfile << substVars(p->opr.op[0]->str.str);
		}
		break;

	case CONNECT:
		if (doConnect() < 0) {
			// error
		}
		break;

	case CREATE: {
		ofstream outfile(substVars(p->opr.op[0]->str.str).c_str());
		}
		break;

	case DISCONNECT:
		if (closeConnection() < 0) {
			// error
		}
		break;

	case ELSETOK:
		if (ex(p->opr.op[0])) {
			return (ex(p->opr.op[1]));
		} else {
			return (ex(p->opr.op[2]));
		}
		break;

	case EQ: return (ex(p->opr.op[0]) == ex(p->opr.op[1]));

	case FILESIZE: {
		ifstream ifile(substVars(p->opr.op[0]->str.str).c_str());
		ifile.seekg(0, ios::end);
		return ifile.tellg();
		}

	case GE: return (ex(p->opr.op[0]) >= ex(p->opr.op[1]));

	case GREP: {
		string result = "";
		string pattern = substVars(p->opr.op[1]->str.str);
		int hasbrace = pattern.find('(');
		if (hasbrace < 0) pattern = '(' + pattern;
		hasbrace = pattern.find(')');
		if (hasbrace < 0) pattern += ')';
		if (grep(pattern, variables[p->opr.op[2]->id.str], result)) {
			variables[p->opr.op[0]->id.str] = result;
		} else {
			variables[p->opr.op[0]->id.str] = "";
		}
		}
		break;

	case HTTP_APPEND:
		if (httpReceive(substVars(p->opr.op[0]->str.str), 0, true) < 0) {
			// error
		}
		break;

	case HTTP_POST:
		if (httpPost(substVars(p->opr.op[0]->str.str)) < 0) {
			// error
		}
		break;

	case HTTP_RECEIVE:
		if (p->opr.nops == 2) {
			if (httpReceive(substVars(p->opr.op[0]->str.str), p->opr.op[1]->con.value) < 0) {
				// error
			}
		} else if (p->opr.nops == 1) {
			if (httpReceive(substVars(p->opr.op[0]->str.str)) < 0) {
				// error
			}
		} else {
			if (httpReceive("") < 0) {
				// error
			}
		}
		break;

	case IF: if (ex(p->opr.op[0])) { return ex(p->opr.op[1]); } break;

	case LINEWISE: doProcessFile(p, true); break;

	case LE: return (ex(p->opr.op[0]) <= ex(p->opr.op[1]));

	case NE: return (ex(p->opr.op[0]) != ex(p->opr.op[1]));

	case PERFORM: return performProcedure(p->opr.op[0]->str.str);

	case PRINT:
		if (p->opr.op[0]->type == typeId) {
			string str =  p->opr.op[0]->id.str;
			variables_type::iterator it = variables.find(str);
			if (it != variables.end()) {
				std::pair<std::string, std::string> pr = *it;
				cout << pr.first << " = " << pr.second << endl;
			} else {
				cout << "Variable " << str << " undefined." << endl;
			}
		} else {
			// print literal string ...
			if (p->opr.nops == 2) {
				// ... to file
				ofstream outfile(substVars(p->opr.op[1]->str.str).c_str());
				outfile << p->opr.op[0]->str.str;
			} else {
				// ... to stdout
				cout << substVars(p->opr.op[0]->str.str);
			}
		}
		break;

	case PROCESS: doProcessFile(p); break;

	case RECEIVE:
		if (p->opr.nops == 2) {
			if (doReceive(substVars(p->opr.op[0]->str.str), p->opr.op[1]->con.value) < 0) {
				// error
			}
		} else if (p->opr.nops == 1) {
			if (doReceive(substVars(p->opr.op[0]->str.str)) < 0) {
				// error
			}
		} else {
			if (doReceive("") < 0) {
				// error
			}
		}
		break;

	case SEND:
		if (sendFile(substVars(p->opr.op[0]->str.str)) < 0) {
			// error
		}
		break;

	case SENDSTRING:
		if (sendString(substVars(p->opr.op[0]->str.str)) < 0) {
			// error
		}
		break;

	case SIZE:
		return (substVars(variables[p->opr.op[0]->id.str]).size());

	case SUBSTVARS:
		switch (p->opr.op[1]->type) {
		case typeStr: variables[p->opr.op[0]->id.str] = substVars(p->opr.op[1]->str.str);
				break;
		case typeId: variables[p->opr.op[0]->id.str] = substVars(variables[p->opr.op[1]->id.str]);
				break;
		}
		break;

	case WHILE:
		while (ex(p->opr.op[0])) {
			ex(p->opr.op[1]);
		}
		return 0;

	case '+': return (ex(p->opr.op[0]) + ex(p->opr.op[1]));
	case '-': return (ex(p->opr.op[0]) - ex(p->opr.op[1]));
	case '<': return (ex(p->opr.op[0]) < ex(p->opr.op[1]));
	case '>': return (ex(p->opr.op[0]) > ex(p->opr.op[1]));
	case '=': doVariableAssignment(p); break;
	case ';': ex(p->opr.op[0]); return ex(p->opr.op[1]); break;
	}
	return 0;
}
